#include <stdio.h>
#include <string.h>

#define KITE_COMPILE_DEFINE
#include "kite_object.h"
#include "kite_vm.h"
#include "kite_opcodes.h"

#define COMPILE_INSTRUCTION(i, j) \
    { \
        kite_thread_t *thd = compiler->thd; \
        kite_object_t *cur = (kite_object_t*)kite_top_stack(compiler->obj_created); \
        COMPILE_INSTRUCTION2(thd, compiler, i, j, &cur->builtin_data.funcvalue.funcptr); \
        (void)cur; \
    }
    
void COMPILE_INSTRUCTION2(kite_thread_t *thd, kite_compiler_t *compiler, kite_opcode_t *inst, int lno, void **dest) 
    { 
        kite_opcode_t *fptr = *dest;
        inst->file = strdup(compiler->file); 
        inst->line = lno; 
        kite_add_to_instruction_list(thd, &fptr, inst);
        *dest = fptr;
    }

static void kite_optimize_increment(kite_compiler_t *compiler, kite_opcode_t *o1, kite_opcode_t *o2)
{
    kite_thread_t *thd = compiler->thd;
    kite_vm_t *vm = thd->vm;
    kite_opcode_t *p1 = (void*)vm->text_segment + (long)o1;
    kite_opcode_t *p2 = (void*)vm->text_segment + (long)o2;
    if (p1->opcode == PUSH && p2->opcode == PUSH) {
        if (p1->op_arg.obj->type == OBJ_IDENT && p2->op_arg.obj->type == OBJ_IDENT) {
            if (strcmp(p1->op_arg.obj->builtin_data.stringvalue.string,
                       p2->op_arg.obj->builtin_data.stringvalue.string) == 0) {
                if ((p1 + 1) < ((void*)vm->text_segment + vm->text_segment_length) &&
                    (p1 + 2) < ((void*)vm->text_segment + vm->text_segment_length)) {
                    if ((p1 + 1)->opcode == DEREF_1 && (p2 + 1)->opcode == DEREF_1) {
                        p2->opcode = DUPE_TOP;
                        p2->op_arg.dupe_ref = TRUE;
                        (p2 + 1)->opcode = NOP;
                    }
                }
            }
        }
    }
}

static void *kite_compile_instruction_break(kite_compiler_t *compiler, int line)
{
    kite_opcode_t *opc = kite_compile_jump_uncond(NULL);
    opc->opcode = TEMP_BREAK;
    COMPILE_INSTRUCTION(opc, line);
    return NULL;
}

static void *kite_compile_instruction_continue(kite_compiler_t *compiler, int line)
{
    kite_opcode_t *opc = kite_compile_jump_uncond(NULL);
    opc->opcode = TEMP_CONTINUE;
    COMPILE_INSTRUCTION(opc, line);
    return NULL;
}

static void *kite_compile_instruction_return(kite_compiler_t *compiler, int line)
{
    kite_opcode_t *opc = kite_compile_return();
    COMPILE_INSTRUCTION(opc, line);
    return NULL;
}

static void *kite_compile_instruction_func_begin(kite_compiler_t *compiler, int line, kite_object_t *param_list)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *method = 
        kite_new_method_bytecode(thd, NULL, 
            kite_list_count(thd, param_list));
            
    /* Bypass declaration so we don't accidentally run the method. */        
    COMPILE_INSTRUCTION(kite_compile_jump_uncond(NULL), line);
    kite_push_stack(&compiler->func_skip_locs, FALSE, compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t));
    
    kite_push_stack(&compiler->obj_created, FALSE,
                    method);
    kite_set_arginfo(method, param_list);
    COMPILE_INSTRUCTION(kite_compile_funcargs(param_list), line);
    return NULL;
}

static void *kite_compile_instruction_func_body(kite_compiler_t *compiler, int line, char *func_name, char *doc_string)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *obj;
    kite_object_t *parent;
    void *pc;
    kite_opcode_t *opc;
    
    COMPILE_INSTRUCTION(kite_compile_return(), line);
    
    obj = kite_pop_stack(&compiler->obj_created);
    parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
    
    obj->parent = kite_reference_object(parent);
    kite_add_method(thd, parent, func_name, obj);
    kite_set_docstring(obj, doc_string);
    
    pc = kite_top_stack(compiler->func_skip_locs);
    if (pc)
    {
        COMPILE_INSTRUCTION(kite_compile_nop(), line);
        opc = (void*)thd->vm->text_segment + (long)pc;
        opc->op_arg.jumpto = thd->vm->cur_text_segment - sizeof(kite_opcode_t);
        kite_pop_stack(&compiler->func_skip_locs);
    }
    
    return NULL;
}

static void *kite_compile_instruction_anon_func_begin(kite_compiler_t *compiler, int line, kite_object_t *param_list)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *method = 
        kite_new_method_bytecode(thd, NULL, 
            kite_list_count(thd,param_list));
    
    /* Bypass declaration so we don't accidentally run the method. */        
    COMPILE_INSTRUCTION(kite_compile_jump_uncond(NULL), line);
    kite_push_stack(&compiler->func_skip_locs, FALSE, compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t));
        
    kite_push_stack(&compiler->obj_created, FALSE,
                    method);
    kite_set_arginfo(method, param_list);
    COMPILE_INSTRUCTION(kite_compile_funcargs(param_list), line);
    return NULL;
}

static void *kite_compile_instruction_anon_func_body(kite_compiler_t *compiler, int line)
{
    kite_object_t *obj;
    kite_opcode_t *opc;
    void *pc;

    COMPILE_INSTRUCTION(kite_compile_return(), line);
    obj = kite_pop_stack(&compiler->obj_created);
    COMPILE_INSTRUCTION(kite_compile_push(obj), line);
    
    pc = kite_top_stack(compiler->func_skip_locs);
    if (pc)
    {
        opc = (void*)compiler->thd->vm->text_segment + (long)pc;
        opc->op_arg.jumpto = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
        kite_pop_stack(&compiler->func_skip_locs);
    }

    return NULL;
}

static void *kite_compile_instruction_class_begin(kite_compiler_t *compiler, int line, char *name)
{
    kite_thread_t *thd = compiler->thd;
    (void)line;
    
    /* Bypass declaration so we don't accidentally run the class. */        
    COMPILE_INSTRUCTION(kite_compile_jump_uncond(NULL), line);
    kite_push_stack(&compiler->func_skip_locs, FALSE, compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t));
    
    kite_push_stack(&compiler->obj_created, FALSE,
                    kite_new_class(thd, NULL, name));
    return NULL;
}

static void *kite_compile_instruction_class_body(kite_compiler_t *compiler, int line, char *name, char *doc_string)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *obj;
    kite_funccall_info_t *fcall;
    kite_object_t *parent;
    kite_opcode_t *opc;
    void *pc;
    
    COMPILE_INSTRUCTION(kite_compile_return(), line);
    obj = kite_pop_stack(&compiler->obj_created);
    
    pc = kite_top_stack(compiler->func_skip_locs);
    if (pc)
    {
        COMPILE_INSTRUCTION(kite_compile_nop(), line);
        opc = (void*)compiler->thd->vm->text_segment + (long)pc;
        opc->op_arg.jumpto = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
        kite_pop_stack(&compiler->func_skip_locs);
    }
    
    obj->builtin_data.funcvalue.functype = FUNC_BYTECODE;
    parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
    kite_add_property(thd, parent, name, 1, NULL);
    kite_set_property(thd, parent, name, obj);
    kite_set_docstring(obj, doc_string);

#ifndef HAVE_GC_H
    fcall = (kite_funccall_info_t*)malloc(sizeof(kite_funccall_info_t));
#else
    fcall = (kite_funccall_info_t*)GC_malloc(sizeof(kite_funccall_info_t));
#endif /* HAVE_GC_H */

    fcall->file = compiler->file;
    fcall->line = line;
    fcall->locals = NULL;
    fcall->sys_or_user = 0;
    fcall->this = obj;
    fcall->last_deref = NULL;
    fcall->last_deref_obj = NULL;
    kite_push_stack(&thd->func_stack, FALSE, fcall);

    kite_vm_call_object(thd, obj, obj, kite_new_list(thd));

    if (thd->exception) return NULL;

    kite_pop_stack(&thd->func_stack);
    kite_destruct_symtab(thd, &fcall->locals, TRUE);
#ifndef HAVE_GC_H
    free(fcall);
#endif /* HAVE_GC_H */

    return NULL;
}

static void *kite_compile_instruction_from_default(kite_compiler_t *compiler, int line)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *obj = (kite_object_t*)kite_top_stack(compiler->obj_created);
    (void)line;
    obj->object_data.inherit_from = kite_dereference_and_load(thd, "System.object");
    return NULL;
}

static void *kite_compile_instruction_from_given(kite_compiler_t *compiler, int line, kite_object_t *from_class)
{
    kite_object_t *obj = (kite_object_t*)kite_top_stack(compiler->obj_created);
    (void)line;
    obj->object_data.inherit_from = from_class;
    assert(obj->object_data.inherit_from->type == OBJ_CLASS ||
           obj->object_data.inherit_from->type == OBJ_INSTANCE);
    return NULL;
}

static void *kite_compile_instruction_immediate_deref_1(kite_compiler_t *compiler, int line, char *name)
{
    #ifndef HAVE_GC_H
            kite_funccall_info_t *fcall = (kite_funccall_info_t*)malloc(sizeof(kite_funccall_info_t));
    #else
            kite_funccall_info_t *fcall = (kite_funccall_info_t*)GC_malloc(sizeof(kite_funccall_info_t));
    #endif /* HAVE_GC_H */
            void *opc = NULL;
            int idx;
            kite_thread_t *thd = compiler->thd;
            int old_offset = thd->vm->cur_text_segment;
            
            COMPILE_INSTRUCTION2(thd, compiler, kite_compile_push(kite_new_ident(thd, name)), 
                                 line, &opc);
            COMPILE_INSTRUCTION2(thd, compiler, kite_compile_deref_1(TRUE), line, &opc);
            COMPILE_INSTRUCTION2(thd, compiler, kite_compile_return(), line, &opc);
            
            if (old_offset == 0)
                opc -= sizeof(kite_opcode_t);
            
            fcall->file = compiler->file;
            fcall->line = line;
            fcall->locals = NULL;
            fcall->last_deref = NULL;
            fcall->last_deref_obj = NULL;
            fcall->sys_or_user = 0;
            idx = compiler->obj_created->length;
            fcall->this =
                (kite_object_t*)compiler->obj_created->stack
                    [idx >= 2 ? idx - 2 : idx - 1].obj;
            kite_push_stack(&thd->func_stack, FALSE, fcall);

            kite_vm_execute_user_method(thd, opc);

            if (thd->exception) return 0;

            kite_pop_stack(&thd->func_stack);
    #ifndef HAVE_GC_H
            free(fcall);
    #endif /* HAVE_GC_H */

            kite_free_instruction_list(thd, opc);
            return kite_vm_pop(thd);
}

static void *kite_compile_instruction_immediate_deref_2(kite_compiler_t *compiler, int line, kite_object_t *obj, char *name)
{
    void *opc = NULL;
    int idx;
    kite_thread_t *thd = compiler->thd;
#ifndef HAVE_GC_H
    kite_funccall_info_t *fcall = (kite_funccall_info_t*)malloc(sizeof(kite_funccall_info_t));
#else
    kite_funccall_info_t *fcall = (kite_funccall_info_t*)GC_malloc(sizeof(kite_funccall_info_t));
#endif /* HAVE_GC_H */
    COMPILE_INSTRUCTION2(thd, compiler, kite_compile_push(obj), 
                         line, &opc);
    COMPILE_INSTRUCTION2(thd, compiler, kite_compile_push(kite_new_ident(thd, name)), 
                         line, &opc);
    COMPILE_INSTRUCTION2(thd, compiler, kite_compile_deref_2(), line, &opc);
    COMPILE_INSTRUCTION2(thd, compiler, kite_compile_return(), line, &opc);
    
    fcall->file = compiler->file;
    fcall->line = line;
    fcall->locals = NULL;
    fcall->last_deref = NULL;
    fcall->last_deref_obj = NULL;
    fcall->sys_or_user = 0;
    idx = compiler->obj_created->length;
    fcall->this =
        (kite_object_t*)compiler->obj_created->stack
            [idx >= 2 ? idx - 2 : idx - 1].obj;
    kite_push_stack(&thd->func_stack, FALSE, fcall);
    
    kite_vm_execute_user_method(thd, opc);
    
    if (thd->exception) return 0;
    
    kite_pop_stack(&thd->func_stack);
#ifndef HAVE_GC_H
    free(fcall);
#endif /* HAVE_GC_H */

    kite_free_instruction_list(thd, opc);
    return kite_vm_pop(thd);
}

static void *kite_compile_instruction_constructor(kite_compiler_t *compiler, int line, kite_object_t *func_args)
{
    kite_thread_t *thd = compiler->thd;
    
    /* Bypass declaration so we don't accidentally run the method. */        
    COMPILE_INSTRUCTION(kite_compile_jump_uncond(NULL), line);
    kite_push_stack(&compiler->func_skip_locs, FALSE, compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t));
    
    kite_push_stack(&compiler->obj_created, FALSE,
                     kite_new_method_bytecode(thd, NULL, 
                                    kite_list_count(thd, func_args)));
    COMPILE_INSTRUCTION(kite_compile_funcargs(func_args), line);
    return NULL;
}

static void *kite_compile_instruction_constructor_body(kite_compiler_t *compiler, int line, kite_object_t *func_args, char *doc_str)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *obj;
    kite_object_t *parent;
    kite_opcode_t *opc;
    void *pc;
    
    COMPILE_INSTRUCTION(kite_compile_return(), line);
    
    obj = kite_pop_stack(&compiler->obj_created);
    parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
    
    obj->parent = kite_reference_object(parent);
    kite_add_method(thd, parent, "__construct__", obj);
    kite_set_docstring(obj, doc_str);
    kite_set_arginfo(obj, func_args);
    
    pc = kite_top_stack(compiler->func_skip_locs);
    if (pc)
    {
        COMPILE_INSTRUCTION(kite_compile_nop(), line);
        opc = (void*)compiler->thd->vm->text_segment + (long)pc;
        opc->op_arg.jumpto = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
        kite_pop_stack(&compiler->func_skip_locs);
    }
    
    return NULL;
}

static void *kite_compile_instruction_destructor(kite_compiler_t *compiler, int line)
{
    kite_thread_t *thd = compiler->thd;
    (void)line;
    
    /* Bypass declaration so we don't accidentally run the method. */        
    COMPILE_INSTRUCTION(kite_compile_jump_uncond(NULL), line);
    kite_push_stack(&compiler->func_skip_locs, FALSE, compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t));
    
    kite_push_stack(&compiler->obj_created, FALSE,
                     kite_new_method_bytecode(thd, NULL, 0));
    return NULL;                
}

static void *kite_compile_instruction_destructor_body(kite_compiler_t *compiler, int line, char *doc_str)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *obj;
    kite_object_t *parent;
    kite_opcode_t *opc;
    void *pc;
    
    COMPILE_INSTRUCTION(kite_compile_return(), line);
    
    obj = kite_pop_stack(&compiler->obj_created);
    parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
    
    pc = kite_top_stack(compiler->func_skip_locs);
    if (pc)
    {
        COMPILE_INSTRUCTION(kite_compile_nop(), line);
        opc = (void*)compiler->thd->vm->text_segment + (long)pc;
        opc->op_arg.jumpto = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
        kite_pop_stack(&compiler->func_skip_locs);
    }
    
    obj->parent = kite_reference_object(parent);
    kite_add_method(thd, parent, "__destruct__", obj);
    kite_set_docstring(obj, doc_str);
    
    return NULL;
}

static void *kite_compile_instruction_property_default(kite_compiler_t *compiler, int line, int global, char *name, char *doc_str)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
    kite_symtab_t *symt;
    
    if (parent->type == OBJ_METHOD) {
        COMPILE_INSTRUCTION(
            kite_compile_push(kite_new_ident(thd, name)), line);
        COMPILE_INSTRUCTION(kite_compile_defprop(), line);
    } else {
        kite_add_property(thd, parent, name, global, NULL);
        KITE_FIND_PROPERTY(symt, parent, name, global);
        symt->docstring = strdup(doc_str);
    }
    
    return NULL;
}

static void *kite_compile_instruction_property_specified(kite_compiler_t *compiler, int line, int global, char *name, char *doc_str)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
    kite_symtab_t *symt;

    COMPILE_INSTRUCTION(
        kite_compile_push(kite_new_ident(thd, name)), line);
    COMPILE_INSTRUCTION(kite_compile_deref_1(FALSE), line);
    if (parent->type == OBJ_METHOD) {
        COMPILE_INSTRUCTION(
            kite_compile_push(kite_new_ident(thd, name)), line);
        COMPILE_INSTRUCTION(kite_compile_defprop(), line);
    } else {
        kite_add_property(thd, parent, name, global, NULL);
        KITE_FIND_PROPERTY(symt, parent, name, global);
        symt->docstring = strdup(doc_str);
    }
    
    return NULL;
}

static void *kite_compile_instruction_property_specified_rhs(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_ASSIGNMENT), line);
    return NULL;
}

static void *kite_compile_instruction_operator(kite_compiler_t *compiler, int line, kite_object_t *func_args)
{
    kite_thread_t *thd = compiler->thd;
    
    /* Bypass declaration so we don't accidentally run the method. */        
    COMPILE_INSTRUCTION(kite_compile_jump_uncond(NULL), line);
    kite_push_stack(&compiler->func_skip_locs, FALSE, compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t));
    
    kite_push_stack(&compiler->obj_created, FALSE,
                     kite_new_method_bytecode(thd, NULL, 
                                    kite_list_count(thd, func_args)));
    COMPILE_INSTRUCTION(kite_compile_funcargs(func_args), line);
    
    return NULL;
}

static void *kite_compile_instruction_operator_body(kite_compiler_t *compiler, int line, int op_type, char *doc_str)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *obj;
    kite_object_t *parent;
    kite_opcode_t *opc;
    void *pc;
        
    COMPILE_INSTRUCTION(kite_compile_return(), line);
    
    obj = kite_pop_stack(&compiler->obj_created);
    parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
    
    kite_add_operator(thd, parent, op_type, obj);
    kite_set_docstring(obj, doc_str);
    
    pc = kite_top_stack(compiler->func_skip_locs);
    if (pc)
    {
        COMPILE_INSTRUCTION(kite_compile_nop(), line);
        opc = (void*)compiler->thd->vm->text_segment + (long)pc;
        opc->op_arg.jumpto = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
        kite_pop_stack(&compiler->func_skip_locs);
    }
    
    return NULL;
}

static void *kite_compile_instruction_assignment_lhs(kite_compiler_t *compiler, int line)
{
    kite_object_t *cur = (kite_object_t*)kite_top_stack(compiler->obj_created);
    kite_opcode_t *opc;
    void *pc = cur->builtin_data.funcvalue.funcptr;
    (void)line;
    
    opc = (void*)compiler->thd->vm->text_segment + (long)pc + cur->builtin_data.funcvalue.length - sizeof(kite_opcode_t);
        
    if (opc->opcode == ARITH_OP && ((kite_opcode_arithop*)opc)->op_arg.operation == OP_ARRAY_DEREF)
        return (void*)opc - (long)compiler->thd->vm->text_segment;
    else
        return NULL;
}

static void *kite_compile_instruction_assignment_rhs(kite_compiler_t *compiler, int line, kite_opcode_t *op1, kite_opcode_t *orig_op2)
{
    kite_opcode_t *op2;
    int has_possible_dupe = TRUE;
    int i;
    kite_vm_t *vm = compiler->thd->vm;
    
    /* Ensure next two instructions are valid for dupe optimization. 
       Equivalent to op1 && op1->next && op1->next->next in old code. */
    op2 = (void*)vm->text_segment + (long)op1;
    for (i = 0; i < 2; i++)
    {
        if (op2->opcode == RETURN_NOW || ((void*)op2 - (long)vm->text_segment) >= vm->text_segment_length)
        {
            has_possible_dupe = FALSE;
            break;
        }
        op2++;
    }
    
    /* Perform dupe optimization. */
    if (has_possible_dupe)
    {
        op2 = (void*)op2 - (long)vm->text_segment;
        kite_optimize_increment(compiler, op1, op2);
    }
    
    if (!orig_op2)
    {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_ASSIGNMENT), line);
    }
    else
    {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_ARRAY_SET), line);
        orig_op2->opcode = NOP;
    }
    
    return NULL;
}

static void *kite_compile_instruction_or_lhs(kite_compiler_t *compiler, int line)
{
    kite_opcode_t *opc = kite_compile_jump_true(NULL);
    COMPILE_INSTRUCTION(kite_compile_dupe_top(FALSE), line);
    COMPILE_INSTRUCTION(opc, line);
    return compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
}

static void *kite_compile_instruction_or_rhs(kite_compiler_t *compiler, int line, kite_opcode_t *lhs_opc)
{
    kite_opcode_t *opc = kite_compile_nop();

    (void)compiler;
    opc->file = "";
    opc->line = line;

    COMPILE_INSTRUCTION(kite_compile_arithop(OP_OR), line);
    COMPILE_INSTRUCTION(opc, line);
    lhs_opc = (void*)lhs_opc + (long)compiler->thd->vm->text_segment;
    ((kite_opcode_jump_true*)lhs_opc)->op_arg.jumpto = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
    
    return NULL;
}

static void *kite_compile_instruction_xor(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_XOR), line);
    return NULL;
}

static void *kite_compile_instruction_and_lhs(kite_compiler_t *compiler, int line)
{
    kite_opcode_t *opc = kite_compile_jump_false(NULL);
    COMPILE_INSTRUCTION(kite_compile_dupe_top(FALSE), line);
    COMPILE_INSTRUCTION(opc, line);
    return compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
}

static void *kite_compile_instruction_and_rhs(kite_compiler_t *compiler, int line, kite_opcode_t *lhs_opc)
{
    kite_opcode_t *opc = kite_compile_nop();

    (void)compiler;
    opc->file = "";
    opc->line = line;

    COMPILE_INSTRUCTION(kite_compile_arithop(OP_AND), line);
    COMPILE_INSTRUCTION(opc, line);
    lhs_opc = (void*)lhs_opc + (long)compiler->thd->vm->text_segment;
    ((kite_opcode_jump_true*)lhs_opc)->op_arg.jumpto = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
    
    return NULL;
}

static void *kite_compile_instruction_left_shift(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_LEFT_SHIFT), line);
    return NULL;
}

static void *kite_compile_instruction_right_shift(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_RIGHT_SHIFT), line);
    return NULL;
}

static void *kite_compile_instruction_equals(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_EQUALS), line);
    return NULL;
}

static void *kite_compile_instruction_not_equals(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_NOT_EQUALS), line);
    return NULL;
}

static void *kite_compile_instruction_is(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_objis(FALSE), line);
    return NULL;
}

static void *kite_compile_instruction_isof(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_objis(TRUE), line);
    return NULL;
}

static void *kite_compile_instruction_less_than(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_LESS_THAN), line);
    return NULL;
}

static void *kite_compile_instruction_greater_than(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_GREATER_THAN), line);
    return NULL;
}

static void *kite_compile_instruction_less_equals(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_LESS_OR_EQUALS), line);
    return NULL;
}

static void *kite_compile_instruction_greater_equals(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_GREATER_OR_EQUALS), line);
    return NULL;
}

static void *kite_compile_instruction_add(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_ADD), line);
    return NULL;
}

static void *kite_compile_instruction_subtract(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_SUBTRACT), line);
    return NULL;
}

static void *kite_compile_instruction_multiply(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_MULTIPLY), line);
    return NULL;
}

static void *kite_compile_instruction_divide(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_DIVIDE), line);
    return NULL;
}

static void *kite_compile_instruction_modulus(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_MODULUS), line);
    return NULL;
}

static void *kite_compile_instruction_unary_plus(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_push(kite_new_null(thd)), line);
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_UNARY_PLUS), line);
    return NULL;
}

static void *kite_compile_instruction_unary_minus(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_push(kite_new_null(thd)), line);
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_UNARY_MINUS), line);
    return NULL;
}

static void *kite_compile_instruction_not(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_push(kite_new_null(thd)), line);
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_NOT), line);
    return NULL;
}

static void *kite_compile_instruction_map(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_MAP), line);
    return NULL;
}

static void *kite_compile_instruction_reduce(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_REDUCE), line);
    return NULL;
}

static void *kite_compile_instruction_func_call(kite_compiler_t *compiler, int line, char *name)
{
	COMPILE_INSTRUCTION(kite_compile_this(), line);
    COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, name)), line);
    COMPILE_INSTRUCTION(kite_compile_deref_1(FALSE), line);
    return NULL;
}

static void *kite_compile_instruction_func_args(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_call(), line);
    return NULL;
}

static void *kite_compile_instruction_deref_and_call_begin(kite_compiler_t *compiler, int line, char *name)
{
    COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, name)), line);
    return NULL;
}

static void *kite_compile_instruction_deref_and_call_args(kite_compiler_t *compiler, int line, kite_opcode_t *func_args)
{
    if (func_args->line == 0) func_args->line = line;
    COMPILE_INSTRUCTION(kite_compile_call(), line);
    return NULL;
}

static void *kite_compile_instruction_array_deref(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_arithop(OP_ARRAY_DEREF), line);
    return NULL;
}

static void *kite_compile_instruction_deref_2(kite_compiler_t *compiler, int line, char *name)
{
    COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, name)), line);
    COMPILE_INSTRUCTION(kite_compile_deref_2(), line);
    return NULL;
}

static void *kite_compile_instruction_deref_1(kite_compiler_t *compiler, int line, char *name)
{
    kite_thread_t *thd = compiler->thd;
    kite_vm_t *vm = thd->vm;
    kite_opcode_t *op = kite_compile_push(kite_new_ident(thd, name));
    COMPILE_INSTRUCTION(op, line);
    op = vm->cur_text_segment - sizeof(kite_opcode_t);
    COMPILE_INSTRUCTION(kite_compile_deref_1(TRUE), line);
    return op;
}

static void *kite_compile_instruction_empty_param_list(kite_compiler_t *compiler, int line)
{
    kite_thread_t *thd = compiler->thd;
    kite_opcode_t *opc = kite_compile_push(kite_new_list(thd));
    (void)line;
    COMPILE_INSTRUCTION(opc, line);
    return opc;
}

static void *kite_compile_instruction_single_param_list(kite_compiler_t *compiler, int line)
{
    kite_thread_t *thd = compiler->thd;
    kite_opcode_t *opc = kite_compile_list_cons_1(thd);
    COMPILE_INSTRUCTION(opc, line);
    return opc;
}

static void *kite_compile_instruction_multi_param_list(kite_compiler_t *compiler, int line)
{
    kite_thread_t *thd = compiler->thd;
    kite_opcode_t *opc = kite_compile_list_cons(thd);
    COMPILE_INSTRUCTION(opc, line);
    return opc;
}

static void *kite_compile_instruction_constant_string(kite_compiler_t *compiler, int line, char *str)
{
    kite_thread_t *thd = compiler->thd;
    kite_opcode_t *opc = kite_compile_push(kite_new_string(thd, str));
    COMPILE_INSTRUCTION(opc, line);
    return opc;
}

static void *kite_compile_instruction_constant_integer(kite_compiler_t *compiler, int line, int val)
{
    kite_thread_t *thd = compiler->thd;
    kite_opcode_t *opc = kite_compile_push(kite_new_integer(thd, val));
    COMPILE_INSTRUCTION(opc, line);
    return opc;
}

static void *kite_compile_instruction_constant_float(kite_compiler_t *compiler, int line, double val)
{
    kite_thread_t *thd = compiler->thd;
    kite_opcode_t *opc = kite_compile_push(kite_new_float(thd, val));
    COMPILE_INSTRUCTION(opc, line);
    return opc;
}

static void *kite_compile_instruction_constant_bool(kite_compiler_t *compiler, int line, int val)
{
    kite_thread_t *thd = compiler->thd;
    kite_opcode_t *opc = kite_compile_push(kite_new_boolean(thd, val));
    COMPILE_INSTRUCTION(opc, line);
    return opc;
}

static void *kite_compile_instruction_constant_null(kite_compiler_t *compiler, int line)
{
    kite_thread_t *thd = compiler->thd;
    kite_opcode_t *opc = kite_compile_push(kite_new_null(thd));
    COMPILE_INSTRUCTION(opc, line);
    return opc;
}

static void *kite_compile_instruction_constant_range(kite_compiler_t *compiler, int line, int range_from, int range_to, int range_step)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *range_class = kite_dereference_and_load(thd, "System.math.range");
    kite_object_t *range_args = kite_new_list(thd);
    kite_object_t *range_object;

    kite_append_list(thd, range_args, kite_new_integer(thd, range_from));
    kite_append_list(thd, range_args, kite_new_integer(thd, range_to));
    kite_append_list(thd, range_args, kite_new_integer(thd, range_step));
    range_object = 
        kite_new_instance_with_constructor(thd, range_class, range_args);
    COMPILE_INSTRUCTION(kite_compile_push(range_object), line);
    kite_dereference_object(range_args);
    kite_dereference_object(range_class);
    kite_dereference_object(range_object);
    return NULL;
}

static void *kite_compile_instruction_constant_regex(kite_compiler_t *compiler, int line, char *regex, int ci)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *param_list = kite_new_list(thd);
    kite_object_t *regex_obj = kite_dereference_and_load(thd, "System.regex");
    
    kite_append_list(thd, param_list, kite_new_string(thd, regex));
    kite_append_list(thd, param_list, kite_new_boolean(thd, ci));
    
    COMPILE_INSTRUCTION(kite_compile_push(regex_obj), line);
    COMPILE_INSTRUCTION(kite_compile_push(param_list), line);
    COMPILE_INSTRUCTION(kite_compile_make(), line);
    
    return NULL;
}

static void *kite_compile_instruction_make(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_make(), line);
    return NULL;
}

static void *kite_compile_instruction_zero_element_list(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_list_cons_0(thd), line);
    return NULL;
}

static void *kite_compile_instruction_one_element_list(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_list_cons_1(thd), line);
    return NULL;
}

static void *kite_compile_instruction_two_element_list(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_list_cons(thd), line);
    return NULL;
}

static void *kite_compile_instruction_decide_begin(kite_compiler_t *compiler, int line)
{
    (void)line;
    kite_push_stack(&compiler->decide_exits, FALSE, kite_compile_nop());
    kite_push_stack(&compiler->decide_begins, FALSE, compiler->thd->vm->cur_text_segment);
    return NULL;
}

static void *kite_compile_instruction_decide_end(kite_compiler_t *compiler, int line)
{
    kite_opcode_t *tmpopc = kite_pop_stack(&compiler->decide_exits);
    COMPILE_INSTRUCTION(tmpopc, line);
    void *pc;
    void *tos = kite_pop_stack(&compiler->decide_begins);
    
    while ((pc = kite_top_stack(compiler->decide_exit_locs)) >= tos)
    {
        tmpopc = (void*)compiler->thd->vm->text_segment + (long)pc;
        tmpopc->op_arg.jumpto = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
        kite_pop_stack(&compiler->decide_exit_locs);
    }
    return NULL;
}

static void *kite_compile_instruction_decide_condition_begin(kite_compiler_t *compiler, int line)
{
    kite_opcode_t *opc = kite_compile_jump_false(NULL);
    COMPILE_INSTRUCTION(opc, line);
    return compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
}

static void *kite_compile_instruction_decide_condition_end(kite_compiler_t *compiler, int line, kite_opcode_t *begin_opc)
{
    kite_opcode_jump_false *c = (kite_opcode_jump_false*)begin_opc;
    kite_opcode_t *obj = kite_compile_nop();
    
    COMPILE_INSTRUCTION(
        kite_compile_jump_uncond(
            kite_top_stack(compiler->decide_exits)
        ),
        line
    );
    
    kite_push_stack(&compiler->decide_exit_locs, FALSE, (void*)compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t));
    
    COMPILE_INSTRUCTION(obj, line);
    c = (void*)c + (long)compiler->thd->vm->text_segment;
    c->op_arg.jumpto = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
    
    return NULL;
}

static void *kite_compile_instruction_eval_begin(kite_compiler_t *compiler, int line)
{
    kite_thread_t *thd = compiler->thd;
    kite_object_t *c = kite_dereference_and_load(thd, "System.vm.compiler");
    COMPILE_INSTRUCTION(kite_compile_push(c), line);
    COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, "compileString")), 
        line);
    return NULL;
}

static void *kite_compile_instruction_eval_end(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_list_cons_1(thd), line);
    COMPILE_INSTRUCTION(kite_compile_call(), line);
    COMPILE_INSTRUCTION(kite_compile_eval(), line);
    return NULL;
}

static void *kite_compile_instruction_loop_wrapper(kite_compiler_t *compiler, int line, kite_opcode_t *opc)
{
    kite_opcode_t *loop_bottom_cont = kite_compile_jump_uncond(opc);
    kite_opcode_t *loop_bottom_end = kite_compile_nop();
    kite_opcode_t *tmpopc;
    
    COMPILE_INSTRUCTION(loop_bottom_cont, line);
    loop_bottom_cont = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
    COMPILE_INSTRUCTION(loop_bottom_end, line);
    loop_bottom_end = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);

    /* Fix jump destination */
    opc = (void*)opc + (long)compiler->thd->vm->text_segment;
    while(opc < (void*)compiler->thd->vm->text_segment + (long)loop_bottom_end)
    {
        if ((opc->opcode == JUMP_TRUE && ((kite_opcode_jump_true*)opc)->op_arg.jumpto == NULL) ||
            (opc->opcode == JUMP_FALSE && ((kite_opcode_jump_false*)opc)->op_arg.jumpto == NULL)) 
        {
            ((kite_opcode_jump_true*)opc)->op_arg.jumpto = loop_bottom_end;
        }
        else if (opc->opcode == TEMP_BREAK)
        {
            opc->opcode = JUMP_UNCOND;
            ((kite_opcode_jump_true*)opc)->op_arg.jumpto = loop_bottom_end;
        }
        else if (opc->opcode == TEMP_CONTINUE)
        {
            opc->opcode = JUMP_UNCOND;
            ((kite_opcode_jump_true*)opc)->op_arg.jumpto = loop_bottom_cont;
        }
        opc++;
    }

    return NULL;
}

static void *kite_compile_instruction_while_begin(kite_compiler_t *compiler, int line)
{
    kite_opcode_t *opc = kite_compile_nop();
    COMPILE_INSTRUCTION(opc, line);
    return compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
}

static void *kite_compile_instruction_while_end(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_jump_false(NULL), line);
    return NULL;
}

static void *kite_compile_instruction_until_begin(kite_compiler_t *compiler, int line)
{
    kite_opcode_t *opc = kite_compile_nop();
    COMPILE_INSTRUCTION(opc, line);
    return compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
}

static void *kite_compile_instruction_until_end(kite_compiler_t *compiler, int line)
{
    COMPILE_INSTRUCTION(kite_compile_jump_true(NULL), line);
    return NULL;
}

static void *kite_compile_instruction_run_begin(kite_compiler_t *compiler, int line)
{
    kite_opcode_t *opc = kite_compile_exception_hndl(NULL);
    COMPILE_INSTRUCTION(opc, line);
    return compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
}

static void *kite_compile_instruction_run_end(kite_compiler_t *compiler, int line)
{
    kite_opcode_t *opc = kite_compile_jump_true(NULL);
    COMPILE_INSTRUCTION(kite_compile_push(kite_new_boolean(thd, TRUE)), line);
    COMPILE_INSTRUCTION(opc, line);
    return compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
}

static void *kite_compile_instruction_catch_begin(kite_compiler_t *compiler, int line, kite_opcode_t *run_opc)
{
    kite_opcode_t *opc = kite_compile_nop();
    COMPILE_INSTRUCTION(opc, line);
    
    run_opc = (void*)compiler->thd->vm->text_segment + (long)run_opc;
    ((kite_opcode_exception_hndl*)run_opc)->op_arg.handler = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
    return NULL;
}

static void *kite_compile_instruction_catch_end(kite_compiler_t *compiler, int line, kite_opcode_t *run_opc)
{
    kite_opcode_t *opc = kite_compile_pop_exception_hndl();
    kite_opcode_t *nop = kite_compile_nop();
    kite_opcode_t *jump;
    COMPILE_INSTRUCTION(kite_compile_push(kite_new_boolean(thd, TRUE)), line);

    COMPILE_INSTRUCTION(kite_compile_jump_true(nop), line);
    jump = (void*)compiler->thd->vm->text_segment + (compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t));

    COMPILE_INSTRUCTION(opc, line);
    run_opc = (void*)compiler->thd->vm->text_segment + (long)run_opc;
    ((kite_opcode_jump_true*)run_opc)->op_arg.jumpto = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
    
    COMPILE_INSTRUCTION(nop, line);
    jump->op_arg.jumpto = compiler->thd->vm->cur_text_segment - sizeof(kite_opcode_t);
    return NULL;
}

/**
 * List of compiler actions.
 */
kite_compiler_action_t kite_compiler_actions[] =
{
    (kite_compiler_action_t)kite_compile_instruction_break,
    (kite_compiler_action_t)kite_compile_instruction_continue,
    (kite_compiler_action_t)kite_compile_instruction_return,
    (kite_compiler_action_t)kite_compile_instruction_func_begin,
    (kite_compiler_action_t)kite_compile_instruction_func_body,
    (kite_compiler_action_t)kite_compile_instruction_anon_func_begin,
    (kite_compiler_action_t)kite_compile_instruction_anon_func_body,
    (kite_compiler_action_t)kite_compile_instruction_class_begin,
    (kite_compiler_action_t)kite_compile_instruction_class_body,
    (kite_compiler_action_t)kite_compile_instruction_from_default,
    (kite_compiler_action_t)kite_compile_instruction_from_given,
    (kite_compiler_action_t)kite_compile_instruction_immediate_deref_1,
    (kite_compiler_action_t)kite_compile_instruction_immediate_deref_2,
    (kite_compiler_action_t)kite_compile_instruction_constructor,
    (kite_compiler_action_t)kite_compile_instruction_constructor_body,
    (kite_compiler_action_t)kite_compile_instruction_destructor,
    (kite_compiler_action_t)kite_compile_instruction_destructor_body,
    (kite_compiler_action_t)kite_compile_instruction_property_default,
    (kite_compiler_action_t)kite_compile_instruction_property_specified,
    (kite_compiler_action_t)kite_compile_instruction_property_specified_rhs,
    (kite_compiler_action_t)kite_compile_instruction_operator,
    (kite_compiler_action_t)kite_compile_instruction_operator_body,
    (kite_compiler_action_t)kite_compile_instruction_assignment_lhs,
    (kite_compiler_action_t)kite_compile_instruction_assignment_rhs,
    (kite_compiler_action_t)kite_compile_instruction_or_lhs,
    (kite_compiler_action_t)kite_compile_instruction_or_rhs,
    (kite_compiler_action_t)kite_compile_instruction_xor,
    (kite_compiler_action_t)kite_compile_instruction_and_lhs,
    (kite_compiler_action_t)kite_compile_instruction_and_rhs,
    (kite_compiler_action_t)kite_compile_instruction_left_shift,
    (kite_compiler_action_t)kite_compile_instruction_right_shift,
    (kite_compiler_action_t)kite_compile_instruction_equals,
    (kite_compiler_action_t)kite_compile_instruction_not_equals,
    (kite_compiler_action_t)kite_compile_instruction_is,
    (kite_compiler_action_t)kite_compile_instruction_isof,
    (kite_compiler_action_t)kite_compile_instruction_less_than,
    (kite_compiler_action_t)kite_compile_instruction_greater_than,
    (kite_compiler_action_t)kite_compile_instruction_less_equals,
    (kite_compiler_action_t)kite_compile_instruction_greater_equals,
    (kite_compiler_action_t)kite_compile_instruction_add,
    (kite_compiler_action_t)kite_compile_instruction_subtract,
    (kite_compiler_action_t)kite_compile_instruction_multiply,
    (kite_compiler_action_t)kite_compile_instruction_divide,
    (kite_compiler_action_t)kite_compile_instruction_modulus,
    (kite_compiler_action_t)kite_compile_instruction_unary_plus,
    (kite_compiler_action_t)kite_compile_instruction_unary_minus,
    (kite_compiler_action_t)kite_compile_instruction_not,
    (kite_compiler_action_t)kite_compile_instruction_map,
    (kite_compiler_action_t)kite_compile_instruction_reduce,
    (kite_compiler_action_t)kite_compile_instruction_func_call,
    (kite_compiler_action_t)kite_compile_instruction_func_args,
    (kite_compiler_action_t)kite_compile_instruction_deref_and_call_begin,
    (kite_compiler_action_t)kite_compile_instruction_deref_and_call_args,
    (kite_compiler_action_t)kite_compile_instruction_array_deref,
    (kite_compiler_action_t)kite_compile_instruction_deref_2,
    (kite_compiler_action_t)kite_compile_instruction_deref_1,
    (kite_compiler_action_t)kite_compile_instruction_empty_param_list,
    (kite_compiler_action_t)kite_compile_instruction_single_param_list,
    (kite_compiler_action_t)kite_compile_instruction_multi_param_list,
    (kite_compiler_action_t)kite_compile_instruction_constant_string,
    (kite_compiler_action_t)kite_compile_instruction_constant_integer,
    (kite_compiler_action_t)kite_compile_instruction_constant_float,
    (kite_compiler_action_t)kite_compile_instruction_constant_bool,
    (kite_compiler_action_t)kite_compile_instruction_constant_null,
    (kite_compiler_action_t)kite_compile_instruction_constant_range,
    (kite_compiler_action_t)kite_compile_instruction_constant_regex,
    (kite_compiler_action_t)kite_compile_instruction_make,
    (kite_compiler_action_t)kite_compile_instruction_zero_element_list,
    (kite_compiler_action_t)kite_compile_instruction_one_element_list,
    (kite_compiler_action_t)kite_compile_instruction_two_element_list,
    (kite_compiler_action_t)kite_compile_instruction_decide_begin,
    (kite_compiler_action_t)kite_compile_instruction_decide_end,
    (kite_compiler_action_t)kite_compile_instruction_decide_condition_begin,
    (kite_compiler_action_t)kite_compile_instruction_decide_condition_end,
    (kite_compiler_action_t)kite_compile_instruction_eval_begin,
    (kite_compiler_action_t)kite_compile_instruction_eval_end,
    (kite_compiler_action_t)kite_compile_instruction_loop_wrapper,
    (kite_compiler_action_t)kite_compile_instruction_while_begin,
    (kite_compiler_action_t)kite_compile_instruction_while_end,
    (kite_compiler_action_t)kite_compile_instruction_until_begin,
    (kite_compiler_action_t)kite_compile_instruction_until_end,
    (kite_compiler_action_t)kite_compile_instruction_run_begin,
    (kite_compiler_action_t)kite_compile_instruction_run_end,
    (kite_compiler_action_t)kite_compile_instruction_catch_begin,
    (kite_compiler_action_t)kite_compile_instruction_catch_end,
    NULL
};